#![feature(ip_from)]
#![feature(async_closure)]
#![warn(clippy::expect_used)]
#![deny(clippy::unwrap_used)]

extern crate html5ever;

use futures_util::StreamExt;

use std::{
    collections::HashSet,
    fs::File,
    io::Read,
    net::{IpAddr, Ipv4Addr},
};

use db::{connect, Website};
use metrics::{counter, gauge};
use metrics_exporter_prometheus::PrometheusBuilder;
use serde::Deserialize;
use surrealdb::{engine::remote::ws::Client, Surreal};
use tokio::{io::{AsyncWriteExt, BufWriter}, task::JoinSet};
use tracing::{debug, error, info, instrument, level_filters::LevelFilter, trace, trace_span, warn};
use tracing_subscriber::{fmt, layer::SubscriberExt, EnvFilter, Layer, Registry};

mod db;
mod filesystem;
mod parser;

const GET_METRIC: &str = "total_gets";
const GET_IN_FLIGHT: &str = "gets_in_flight";
const SITES_CRAWLED: &str = "pages_crawled";
const BEING_PROCESSED: &str = "pages_being_processed";

#[derive(Deserialize)]
struct Config {
    surreal_ns: String,
    surreal_db: String,
    surreal_url: String,
    surreal_username: String,
    surreal_password: String,

    crawl_filter: String,
    start_url: String,
    budget: usize,
    batch_size: usize,
}

#[tokio::main]
async fn main() {
    println!("Logs and metrics are provided to the Grafana dashboard");

    let writer = std::fs::OpenOptions::new()
        .append(true)
        .create(true)
        .open("./docker/logs/tracing.log")
        .expect("Couldn't make log file!");

    let filter = EnvFilter::builder()
        .with_default_directive(LevelFilter::DEBUG.into())
        .from_env_lossy();

    let registry = Registry::default().with(
        fmt::layer()
            .with_line_number(true)
            .with_thread_ids(true)
            .with_file(true)
            .json()
            .with_writer(writer)
            .with_filter(filter)
    );

    tracing::subscriber::set_global_default(registry).expect("Failed to set default subscriber");

    let builder = PrometheusBuilder::new();
    builder
        .with_http_listener(std::net::SocketAddr::new(
            IpAddr::V4(Ipv4Addr::from_octets([0, 0, 0, 0])),
            2500,
        ))
        .install()
        .expect("failed to install recorder/exporter");

    info!("Starting...");

    // When getting uncrawled pages, name must contain this variable. "" will effectively get ignored.
    // let crawl_filter = "en.wikipedia.org/";
    // let budget = 50;
    let mut crawled = 0;

    let mut file = File::open("./Crawler.toml").expect("Failed to read Crawler.toml");
    let mut buf = String::new();
    let _ = file.read_to_string(&mut buf);

    let config: Config = toml::from_str(&buf).expect("Failed to parse Crawler.toml");
    let starting_url = &config.start_url;

    let db = connect(&config)
        .await
        .expect("Failed to connect to surreal, aborting.");

    let reqwest = reqwest::Client::builder()
        // .use_rustls_tls()
        .gzip(true)
        .build()
        .expect("Failed to build reqwest client.");

    // Kick off the whole machine - This Website object doesn't matter, it's just to allow for
    // get() to work.
    let span = trace_span!("Pre-Loop");
    let pre_loop_span = span.enter();
    // Download the site
    let site = Website::new(starting_url, false);
    process(site, db.clone(), reqwest.clone()).await;

    drop(pre_loop_span);

    let span = trace_span!("Loop");
    let span = span.enter();
    while crawled < config.budget {
        let uncrawled =
            get_uncrawled_links(&db, config.budget - crawled, config.crawl_filter.clone(), &config).await;
        if uncrawled.is_empty() {
            info!("Had more budget but finished crawling everything.");
            return;
        }

        {
            let mut futures = JoinSet::new();
            for site in uncrawled {
                gauge!(BEING_PROCESSED).increment(1);
                futures.spawn(process(site, db.clone(), reqwest.clone()));
                // let percent = format!("{:.2}%", (crawled as f32 / budget as f32) * 100f32);
                // info!("Crawled {crawled} out of {budget} pages. ({percent})");
            }

            let c = counter!(SITES_CRAWLED);
            // As futures complete runs code in while block
            while futures.join_next().await.is_some() {
                c.increment(1);
                gauge!(BEING_PROCESSED).decrement(1);
                crawled += 1;
            }
        }
    }
    drop(span);

    if let Ok(mut ok) = db
        .query("count(select id from website where crawled = true)")
        .await
    {
        let res = ok.take::<Option<usize>>(0);
        if let Ok(Some(n)) = res {
            info!("Total crawled pages now equals {n}");
        }
    }

    info!("Done");
}

#[instrument(skip(db, reqwest))]
/// Downloads and crawls and stores a webpage.
/// It is acceptable to clone `db`, `reqwest`, and `s3` because they all use `Arc`s internally. - Noted by Oliver
async fn process(mut site: Website, db: Surreal<Client>, reqwest: reqwest::Client) {
    // METRICS
    trace!("Process: {}", &site.site);
    // Build the request
    let request_builder = reqwest.get(site.site.to_string());

    // METRICS
    let g = gauge!(GET_IN_FLIGHT);
    g.increment(1);

    // Send the http request (get)
    if let Ok(response) = request_builder.send().await {
        let headers = response.headers();
        let code = response.status();

        #[allow(non_snake_case)]
        let CT = headers.get("Content-Type");
        let ct = headers.get("content-type");

        if CT.is_none() && ct.is_none() {
        }
        let ct = match (CT,ct) {
            (None, None) => {
                warn!("Server did not respond with Content-Type header. Url: {} Headers: ({:?})", site.site.to_string(), headers);
                return
            },
            (None, Some(a)) => a,
            (Some(a), None) => a,
            (Some(a), Some(_)) => a,
        };

        let path = filesystem::as_path(&site.site, ct);

        // make sure that the file is good to go
        if let Some(file) = filesystem::init(&path).await {
            let should_parse = path.to_string_lossy().ends_with(".html");
            let mut buf: Vec<u8> = Vec::new();

            let mut writer = BufWriter::new(file);

            // Get body from response
            // stream the response onto the disk
            let mut stream = response.bytes_stream();

            info!("Writing at: {:?}", path);
            while let Some(data) = stream.next().await {
                match data {
                    Ok(data) => {
                        let _ = writer.write_all(&data).await;
                        // If we are going to parse this file later, we will save it
                        // into memory as well as the disk.
                        if should_parse {
                            data.iter().for_each(|f| buf.push(*f));
                        }
                    },
                    Err(err) => {
                        eprintln!("{}", err)
                    },
                }
            }
            let _ = writer.flush();

            if should_parse {
                // Parse document and get relationships
                let sites = parser::parse(&site, &buf).await;
                // De-duplicate this list
                let prev_len = sites.len();
                let set = sites.into_iter().fold(HashSet::new(), |mut set, item| {
                    set.insert(item);
                    set
                });
                let de_dupe_sites: Vec<Website> = set.into_iter().collect();
                let diff = prev_len - de_dupe_sites.len();
                trace!("Saved {diff} from being entered into the db by de-duping");
                // Store all the other sites so that we can link to them.
                let _ = Website::store_all(de_dupe_sites, &db).await;
            }

            // METRICS
            g.decrement(1);
            counter!(GET_METRIC).increment(1);

            // update self in db
            site.crawled = true;
            site.status_code = code.as_u16();
            Website::store_all(vec![site], &db).await;
        } else {
            error!("File failed to cooperate: {:?}", path);
        }
    } else {
        error!("Failed to get: {}", &site.site);
    }
}

/// Returns uncrawled links
#[instrument(skip(db, config))]
async fn get_uncrawled_links(
    db: &Surreal<Client>,
    mut count: usize,
    filter: String,
    config: &Config,
) -> Vec<Website> {
    if count > config.batch_size {
        count = config.batch_size;
    }

    debug!("Getting uncrawled links");

    let mut response = db
        .query("SELECT * FROM website WHERE crawled = false AND site ~ type::string($format) LIMIT $count;")
        .bind(("format", filter))
        .bind(("count", count))
        .await
        .expect("Hard-coded query failed..?");
    response
        .take(0)
        .expect("Returned websites couldn't be parsed")
}
